#include <stdlib.h>
#include <stdio.h>
#include "TerrGen.h"

void TerrGen::Subdivide(int topNdx, int botNdx, float newRange, float base, int curDepth) {
   int midNdx = topNdx + (botNdx - topNdx) / 2;
   int rNdx, tNdx, lNdx, bNdx, tr, bl;
   tr = topNdx + 2 * curDepth;
   bl = botNdx - 2 * curDepth;

   if (!curDepth)
      return;

   lNdx = midNdx - curDepth;
   rNdx = midNdx + curDepth;
   tNdx = midNdx - curDepth * side;
   bNdx = midNdx + curDepth * side;

   (*pVec)[tNdx]->y = ((*pVec)[topNdx]->y + (*pVec)[tr]->y) / 2;
   (*pVec)[bNdx]->y = ((*pVec)[bl]->y + (*pVec)[botNdx]->y) / 2;
   (*pVec)[lNdx]->y = ((*pVec)[topNdx]->y + (*pVec)[bl]->y) / 2;
   (*pVec)[rNdx]->y = ((*pVec)[tr]->y + (*pVec)[botNdx]->y) / 2;
  

   (*pVec)[midNdx]->y = ((*pVec)[topNdx]->y + (*pVec)[botNdx]->y + (*pVec)[tr]->y + (*pVec)[bl]->y) / 4;
   
   (*pVec)[lNdx]->y += (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * .01 * newRange;
   (*pVec)[rNdx]->y += (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * .01 * newRange;
   (*pVec)[tNdx]->y += (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * .01 * newRange;
   (*pVec)[bNdx]->y += (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * .01 * newRange;

   (*pVec)[midNdx]->y += (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * newRange;

   Subdivide(topNdx, midNdx, 0.5 * newRange, base, curDepth / 2);
   Subdivide(topNdx + curDepth, midNdx + curDepth, 0.5 * newRange, base, curDepth / 2);
   Subdivide(midNdx - curDepth, botNdx - curDepth, 0.5 * newRange, base, curDepth / 2);
   Subdivide(midNdx, botNdx, 0.5 * newRange, base, curDepth / 2);
}

void TerrGen::Generate() {
   /* Preliminary setup */
   side = (1 << depth) + 1;
   int numPoints = side * side;
   float interval = length / (side - 1);
   float startPos = -length / 2;
   pVec->reserve(numPoints);
   srand(seed);

   /* Initialize the points */
   for (int rNdx = 0; rNdx < side; rNdx++) {
      for (int cNdx = 0; cNdx < side; cNdx++) {
         pVec->push_back(new Point(cNdx * interval + startPos, 0, rNdx * interval + startPos));
      }
   }
   
   //(*pVec)[0]->y = (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * range;
   //(*pVec)[side - 1]->y = (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * range;
   //(*pVec)[numPoints - side]->y = (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * range;
   //(*pVec)[numPoints - 1]->y = (1 - 2 * (rand() & 1)) * (rand() / (float) RAND_MAX) * range;


   //(*pVec)[(numPoints - 1) / 2]->y = ((*pVec)[0]->y + (*pVec)[side - 1]->y + (*pVec)[side * (side - 1)]->y + (*pVec)[numPoints - 1]->y) / 4;

   /* Start the division */
   Subdivide(0, numPoints - 1, range, 0, side / 2);
}
