#include "Vector.h"
#include <iostream>
#include <math.h>

Vector::Vector() {
   x = 0;
   y = 0;
   z = 0;
}

Vector::Vector(float _x, float _y, float _z) {
   x = _x;
   y = _y;
   z = _z;
}

Vector::Vector(const Vector &src) {
   x = src.x;
   y = src.y;
   z = src.z;
}

Vector::Vector(const Point &p1, const Point &p2) {
   x = p1.x - p2.x;
   y = p1.y - p2.y;
   z = p1.z - p2.z;
}

Vector Vector::Normalize() {
   float l = length();
   x /= l;
   y /= l;
   z /= l;
   return *this;
}

Vector Vector::operator= (const Vector &rhs) {
   x = rhs.x;
   y = rhs.y;
   z = rhs.z;
   return *this;
}


Vector Vector::operator= (const float *arr) {
   x = arr[0];
   y = arr[1];
   z = arr[2];
   return *this;
}

Vector Vector::operator+ (const Vector &rhs) const {
   Vector tmp;
   tmp.x = (x + rhs.x);
   tmp.y = (y + rhs.y);
   tmp.z = (z + rhs.z);
   return tmp;
}

Vector Vector::operator- (const Vector &rhs) const {
   Vector tmp;
   tmp.x = (x + rhs.x);
   tmp.y = (y + rhs.y);
   tmp.z = (z + rhs.y);
   return tmp;
}

float Vector::operator* (const Vector &rhs) const {
   return (x * rhs.x) + (y * rhs.y) + (z * rhs.z);
}

float Vector::along(const Vector &rhs) const {
   return (x * rhs.x + y * rhs.y + z * rhs.z) / rhs.length();
}

Vector Vector::operator* (float scalar) const {
   Vector tmp;
   tmp.x = x * scalar;
   tmp.y = y * scalar;
   tmp.z = z * scalar;
   return tmp;
}

int Vector::operator== (const Vector &rhs) const {
   return (x == rhs.x && y == rhs.y && z == rhs.z);
}

Vector Vector::operator/ (float scalar) const {
   Vector tmp;
   tmp.x = x / scalar;
   tmp.y = y / scalar;
   tmp.z = z / scalar;
   return tmp;
}

float Vector::length() const {
   return (sqrt(x * x + y * y + z * z));
}

Vector Vector::genPerp() const {
   Vector tmp;
   tmp.x = (-1)*y;
   tmp.y = x;
   tmp.z = z;
   return tmp;
}
   
Vector Vector::cross(const Vector &rhs) const {
   Vector tmp;
   tmp.x = y * rhs.z - z * rhs.y;
   tmp.y = z * rhs.x - x * rhs.z;
   tmp.z = x * rhs.y - y * rhs.x;
   return tmp;
}

void Vector::setXYZ(float _x, float _y, float _z) {
   x = _x;
   y = _y;
   z = _z;
}

std::ostream &operator<<(std::ostream &os, const Vector &vector)
{
   os << '<' << vector.x << ',' << vector.y << ',' << vector.z << '>';
   return os;
}
