#ifndef HMC_VECTOR
#define HMC_VECTOR
#include <iostream>
#include "Point.h"

struct Point;

struct Vector {
   union {
      struct {
         float x, y, z;
      };
      struct {
         float xyz[3];
      };
   };
   
   Vector();
   Vector(float _x, float _y, float _z);
   Vector(const Vector &src);
   Vector(const Point &p1, const Point &p2);

   Vector operator= (const Vector &rhs);
   Vector operator= (const float *arr);
   float operator* (const Vector &rhs) const;
   float along(const Vector &rhs) const;
   Vector operator* (float scalar) const;
   Vector operator+ (const Vector &rhs) const;
   Vector operator- (const Vector &rhs) const;
   int operator== (const Vector &rhs) const;
   Vector operator/ (float scalar) const;

   float length() const;
   Vector genPerp() const;
   Vector Normalize();

   void setXYZ(float _x, float _y, float _z);
   Vector cross(const Vector &rhs) const;
   
   friend std::ostream &operator<<(std::ostream &os, const Vector &vector);
};
#endif
