#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <vector>
#include <list>
#include <GL/gl.h>
#include <GL/glut.h>
#include "Point.h"
#include "Vector.h"
#include "Triangle.h"
#include "Rectangle.h"
#include "TerrGen.h"
#include "Projectile.h"

#define PI 3.14159
#define MOVE_SCALE .05
#define TERRAIN 256
#define INIT_ENERGY .005

using namespace std;

vector<Point *> pVec;
vector<Triangle *> tVec;
vector<Vector> nVec;
list<Projectile> bList;
Point eye, look;
int GH, GW;
int moveTrigger, buildTrigger;
int side;
float interval, length;
int pointSwt, normalSwt, wireframe, closest;
int vehNdx, vehDir, flyMode;
float phi, theta;
float eyeX, eyeY, eyeZ;
float lookX, lookY, lookZ;
GLUquadricObj* barrelCylinder;

void genNormal(int i); 
void genNormals(); 
int getClosestPoint(Point here);

/* ------------------
 * | Lighting Stuff |
 * ------------------ */

typedef struct materialStruct {
   float ambient[4];
   float diffuse[4];
   float specular[4];
   float shininess[1];
} materialStruct;

materialStruct RedFlat = {
   {0.3, 0.0, 0.0, 1.0},
   {0.9, 0.0, 0.0, 1.0},
   {0.0, 0.0, 0.0, 1.0},
   {0.0}
};

materialStruct GreenShiny = {
   {0.0, 0.3, 0.0, 1.0},
   {0.0, 0.9, 0.0, 1.0},
   {0.2, 1.0, 0.2, 1.0},
   {8.0}
};

materialStruct Opal = {
   {0.8, 0.8, 0.8, 1.0},
   {0.8, 0.8, 0.8, 1.0},
   {1.0, 1.0, 1.0, 1.0},
   {15.0}
};

materialStruct Ground = {
   {0.1, 0.3, 0.1, 1.0},
   {0.1, 0.3, 0.3, 1.0},
   {0.1, 0.1, 0.1, 1.0},
   {0.4}
};

materialStruct Shell = {
   {0.3, 0.3, 0.3, 1.0},
   {0.2, 0.2, 0.2, 1.0},
   {0.4, 0.4, 0.4, 1.0},
   {5.0}
};

float light_pos[4] = {0.0, 1.0, 0.0, 0.0};
float light_amb[4] = {0.2, 0.3, 0.2, 1.0};
float light_diff[4] = {0.2, 0.3, 0.2, 1.0};
float light_spec[4] = {0.4, 0.4, 0.4, 1.0};

void materials(materialStruct materials) {
   glMaterialfv(GL_FRONT, GL_AMBIENT, materials.ambient);
   glMaterialfv(GL_FRONT, GL_DIFFUSE, materials.diffuse);
   glMaterialfv(GL_FRONT, GL_SPECULAR, materials.specular);
   glMaterialfv(GL_FRONT, GL_SHININESS, materials.shininess);
}

void materialHeight(float y) {
   materialStruct tmp = {
      {0.8, 0.4, 0.8, 1.0},
      {0.8, 0.4, 0.8, 1.0},
      {0.2, 0.2, 0.2, 1.0},
      {1.0}
   };
   
   //tmp.diffuse[1] = y;
   tmp.diffuse[0] *= 2 * y;
   tmp.diffuse[2] *= 2 * y;
   
   //tmp.ambient[1] = y;
   tmp.ambient[0] *= 2 * y;
   tmp.ambient[2] *= 2 * y;
   materials(tmp);
}


void init_lighting() {
   glEnable(GL_LIGHT0);
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diff);
   glLightfv(GL_LIGHT0, GL_AMBIENT, light_amb);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light_spec);
}

void pos_light() {
   glMatrixMode(GL_MODELVIEW);
   glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
}

/* ---------------------
 * | Deformation Funcs |
 * --------------------- */

void deformPoint(Vector incid, float energy, int pNdx) {
   float toGive = .25 * energy;
   if (energy == 0)
      return;
  
   if (energy < INIT_ENERGY)
      toGive = 0;

   *pVec[pNdx] = *pVec[pNdx] + incid;
      
   if (pNdx % side) {
      deformPoint(incid, toGive, pNdx - 1);
      if (pNdx >= side) {
         deformPoint(incid, toGive, pNdx - side - 1);
      }
   }

   if (pNdx >= side) {
      deformPoint(incid, toGive, pNdx - side);
      if (pNdx % side != side - 1) {
         deformPoint(incid, toGive, pNdx - side + 1);
      }
   }

   if (pNdx % side != side - 1) {
      deformPoint(incid, toGive, pNdx + 1);
      if (pNdx < (side * (side - 1))) {
         deformPoint(incid, toGive, pNdx + side + 1);
      }
   }

   if (pNdx < (side * (side - 1))) {
      deformPoint(incid, toGive, pNdx + side);
      if (pNdx % side) {
         deformPoint(incid, toGive, pNdx + side - 1);
      }
   }
   genNormal(pNdx);
}

void fireProjectile(Vector aim, Point pos) {
   bList.push_back(Projectile(aim, pos, Vector(0, -.000015, 0), 2 * interval, 5));   
}

void mouseClick(int button, int state, int x, int y) {
   if (GLUT_DOWN == state) {
      if (button == GLUT_LEFT_BUTTON) {
         fireProjectile(Vector(look.x - eye.x, look.y - eye.y, look.z - eye.z).Normalize() * .005, eye);
      }
   }
}

int checkCollision(list<Projectile>::iterator itr) {
   for (int i = 0; i < pVec.size(); i++) {
      if (Vector(itr->pos, *pVec[i]).length() < itr->radius) { // Contact
         printf("Collision Detected\n");
         deformPoint(Vector(*pVec[i], itr->pos), INIT_ENERGY * itr->mass, i);
         buildTrigger = 1;
         return 1;
      }
   }
   
   /*
   for (int i = 0; i < tVec.size(); i++) {
      printf("The dot product is %f\n", fabs(Vector(itr->pos, *(tVec[i]->c)).along(tVec[i]->normal)));
      printf("radius is %f\n", itr->radius);

      if (fabs(Vector(itr->pos, *(tVec[i]->c)).along(tVec[i]->normal)) < itr->radius) { 
         printf("Collision Detected -- triangle %d\n", i);
         //deformPoint(Vector(*pVec[i], itr->pos), INIT_ENERGY * itr->mass, i);
         buildTrigger = 1;
         return 1;
      }
   }
   */
   return 0;
}

void updateProjectiles() {
   for (list<Projectile>::iterator itr = bList.begin(); itr != bList.end(); itr++) {
      itr->Update();
      if (itr->updates > 10000) {
         itr = bList.erase(itr);
      }
      else {
         if (checkCollision(itr))
            itr = bList.erase(itr);
      }
   }
}

/* --------------------
 * | Camera Functions |
 * -------------------- */

void mouseMotion(int x, int y) {  
   float dX, dY;
   dX = (x - GW / 2.0) / (float) GW;
   dY = (y - GH / 2.0) / (float) GH;
   
   theta -= asin(dX);
   if (theta < -PI) {
      theta += 2 * PI;
   }
   else if (theta > PI) {
      theta -= 2 * PI;
   }
   phi -= asin(dY);
   if (phi > PI / 2) {
      phi = PI / 2;
   }
   else if (phi < -PI / 2) {
      phi = - PI / 2;
   }
}

void goDir(int dir) {
   switch (dir) {
   case 0:
      if (vehNdx >= side)
         vehNdx -= side;
      break;
   case 1:
      if (vehNdx >= side && vehNdx % side != side - 1)
         vehNdx -= (side - 1);
      break;
   case 2:
      if (vehNdx % side != side - 1)
         vehNdx++;
      break;
   case 3:
      if (vehNdx % side != side - 1 && vehNdx < (side * (side - 1)))
         vehNdx += side + 1;
      break;
   case 4:
      if (vehNdx < (side * (side - 1)))
         vehNdx += side;
      break;
   case 5:
      if (vehNdx < (side * (side - 1)) && vehNdx % side)
         vehNdx += (side - 1);
      break;
   case 6:
      if (vehNdx % side)
         vehNdx--;
      break;
   case 7:
      if (vehNdx % side && vehNdx >= side)
         vehNdx -= (side + 1);
      break;
   default:
      break;
   }
}

void keyboard(unsigned char key, int x, int y) {
   switch(key) {
   case 'a':
      moveTrigger = 1;
      break;
   case 'd':
      moveTrigger = 2;
      break;
   case 'w':
      moveTrigger = 3;
      break;
   case 's':
      moveTrigger = 4;
      break;
   case 'q':
      exit(0);
      break;
   case 'h':
      printf("Movement controls:\n");
      printf("w -- move forward\n");
      printf("s -- move backward\n");
      printf("a -- strafe left\n");
      printf("d -- strafe right\n");
      printf("Other controls:\n");
      printf("p -- draw points\n");
      printf("o -- draw wireframe\n");
      printf("n -- draw normals\n");
      printf("q -- quit\n");
      printf("h -- this help\n");
      break;
   case 'p':
      pointSwt = !pointSwt;
      break;
   case 'n':
      normalSwt = !normalSwt;
      break;
   case 'o':
      wireframe = !wireframe;
      if (wireframe) {
         glPolygonMode(GL_FRONT, GL_LINE);
         glPolygonMode(GL_BACK, GL_LINE);
      }
      else {
         glPolygonMode(GL_FRONT, GL_FILL);
         glPolygonMode(GL_BACK, GL_FILL);
      }
      break;
   case 'f':
      flyMode = !flyMode;
      break;
   default:
      break;
   }
   closest = getClosestPoint(eye);
}

void idle() {
   if (flyMode) {
      switch   (moveTrigger) {
      case 1:
         eye.x += MOVE_SCALE * sin(theta + PI / 2);
         eye.z += MOVE_SCALE * cos(theta + PI / 2);
         break;
      case 2:
         eye.x -= MOVE_SCALE * sin(theta + PI / 2);
         eye.z -= MOVE_SCALE * cos(theta + PI / 2);
         break;
      case 3:
         eye.z += MOVE_SCALE * cos(theta);
         eye.x += MOVE_SCALE * sin(theta);
         eye.y += MOVE_SCALE * sin(phi);
         break;
      case 4:
         eye.z -= MOVE_SCALE * cos(theta);
         eye.x -= MOVE_SCALE * sin(theta);
         eye.y -= MOVE_SCALE * sin(phi);
         break;
      default:
         break;
      }

   } else {
      switch (moveTrigger) {
      case 1:
         vehDir += 7;
         vehDir = vehDir % 8;
         break;
      case 2:
         vehDir += 1;
         vehDir = vehDir % 8;
         break;
      case 3:
         goDir((vehDir + 4) % 8);
         break;
      case 4:
         goDir(vehDir);
         break;
      default:
         break;
      }
      eye = *pVec[vehNdx];
      eye.y += .07;

   }
   moveTrigger = 0;
  
      
   look.x = eye.x + 2 * sin(theta);
   look.y = eye.y + sin(phi);
   look.z = eye.z + 2 * cos(theta);
   
   updateProjectiles();
   glutWarpPointer(GW / 2, GH / 2);
   glutPostRedisplay();
}

/* -------------------
 * | Misc. Functions |
 * ------------------- */

void reshape(int w, int h) {
   GW = w;
   GH = h;

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   
   gluPerspective(100.0, w / h, .001, 5);
   glMatrixMode(GL_MODELVIEW);
   glViewport(0, 0, w, h);

   glutPostRedisplay();
}

int getClosestPoint(Point here) {
   int curNdx = 0;
   float curLen = INT_MAX;
   for (int i = 0; i < pVec.size(); i++) {
      if (Vector(here, *pVec[i]).length() < curLen) {
         curNdx = i;
      }
   }
   return curNdx;
}

void genTriangles() {
   for (int i = 0; i < side - 1; i++) {
      for (int j = 0; j < side - 1; j++) {
         tVec.push_back(new Triangle(pVec[i * side + j + 1], pVec[i * side + j], pVec[i * side + j + side]));
         tVec.back()->GenNormal();
         tVec.back()->normal.Normalize();
         tVec.push_back(new Triangle(pVec[i * side + j + side + 1], pVec[i * side + j + 1], pVec[i * side + j + side]));
         tVec.back()->GenNormal();
         tVec.back()->normal.Normalize();
      }
   }
}

void genNormal(int i) {
   Vector edges[8];
   Vector tmpVec;
   int count;
   
      count = 0;
      tmpVec = Vector(0, 0, 0);
      
      if (i % side) {
         edges[count++] = Vector(*pVec[i - 1], *pVec[i]).Normalize();
         if (i >= side) {
            edges[count++] = Vector(*pVec[i - 1 - side], *pVec[i]).Normalize();
         }
      }

      if (i >= side) {
         edges[count++] = Vector(*pVec[i - side], *pVec[i]).Normalize();
         if (i % side != side - 1) {
            edges[count++] = Vector(*pVec[i - side + 1], *pVec[i]).Normalize();
         }
      }
      
      if (i % side != side - 1) {
         edges[count++] = Vector(*pVec[i + 1], *pVec[i]).Normalize();
         if (i < (side * (side - 1))) {
            edges[count++] = Vector(*pVec[i + 1 + side], *pVec[i]).Normalize();
         }
      }
      
      if (i < (side * (side - 1))) {
         edges[count++] = Vector(*pVec[i + side], *pVec[i]).Normalize();
         if (i % side) {
            edges[count++] = Vector(*pVec[i + side - 1], *pVec[i]).Normalize();
         }
      }
      for (int j = 0; j < count - 1; j++) {
         tmpVec = tmpVec + (edges[j + 1].cross(edges[j]));
      }
      tmpVec = tmpVec + (edges[0].cross(edges[count - 1]));
      
      tmpVec.Normalize();
      nVec[i] = tmpVec; 
}

void genNormals() {
   nVec.reserve(side * side);
   for (int i = 0; i < pVec.size(); i++) {
      genNormal(i);
   }
}

/* ---------------------
 * | Drawing Functions |
 * --------------------- */

void drawNormals() {
   glBegin(GL_LINES);
   for (int i = 0; i < pVec.size(); i++) {
      glVertex3fv(pVec[i]->xyz);
      glVertex3fv(((*pVec[i]) + nVec[i]).xyz);
   }
   glEnd();
}

void drawTriangles() {
   if (buildTrigger) {
      glDeleteLists(TERRAIN, 1);
      glNewList(TERRAIN, GL_COMPILE_AND_EXECUTE);
      glPushAttrib(GL_CURRENT_BIT); 
      for (int i = 0; i < side - 1; i++) {
         glBegin(GL_TRIANGLE_STRIP);
         for (int j = 0; j < side; j++) {
            glNormal3fv(nVec[i * side + j].xyz);
            materialHeight(pVec[i * side + j]->y);
            glVertex3fv(pVec[i * side + j]->xyz);

            glNormal3fv(nVec[i * side + j + side].xyz);
            materialHeight(pVec[i * side + j + side]->y);
            glVertex3fv(pVec[i * side + j + side]->xyz);
         }
         glEnd();
      }
      glPopAttrib();
      glEndList();
      buildTrigger = 0;
   }
   else {
      glCallList(TERRAIN);
   }
}

void drawBadTriangles() {
   materials(Ground);
   glShadeModel(GL_FLAT);
   glBegin(GL_TRIANGLES);
   for (int i = 0; i < tVec.size(); i++) {
      glNormal3fv(tVec[i]->normal.xyz);
      glVertex3fv(tVec[i]->a->xyz);
      glVertex3fv(tVec[i]->b->xyz);
      glVertex3fv(tVec[i]->c->xyz);
   }
   glEnd();
}

void drawBadNormals() {
   glBegin(GL_LINES);
   for (int i = 0; i < tVec.size(); i++) {
      glVertex3fv(tVec[i]->a->xyz);
      glVertex3fv((*tVec[i]->a + tVec[i]->normal).xyz);
   }
   glEnd();
}

void drawPoints() {
   glBegin(GL_POINTS);
   for (int i = 0; i < pVec.size(); i++) {
      glVertex3fv(pVec[i]->xyz);
   }
   glEnd();
}

void drawProjectiles() {
   materials(Shell);
   for (list<Projectile>::iterator itr = bList.begin(); itr != bList.end(); itr++) {
      glPushMatrix();
      glTranslatef(itr->pos.x, itr->pos.y, itr->pos.z);   
      glutSolidSphere(interval / 2, 10, 10);
      glPopMatrix();
   }
}

void drawTank() {
   glPushMatrix(); {
      glTranslatef(pVec[vehNdx]->x, pVec[vehNdx]->y, pVec[vehNdx]->z);
      glPushMatrix(); {
         materials(Shell);
         glRotatef(-45 * vehDir, 0, 1, 0);
         glScalef(0.5, 0.5, 1);
         glutSolidCube(.1);
      } glPopMatrix();
      
      glPushMatrix(); {
         if (!flyMode) {
            glRotatef(180.0 / PI * theta, 0, 1, 0);
         }
         glPushMatrix(); {
            glTranslatef(0.0, .025, 0.0);
            glutSolidSphere(.025, 10, 10);
         } glPopMatrix();

         glPushMatrix(); {
            glTranslatef(0.0, 0.025, 0.0);
            glRotatef(-35, 1, 0, 0);
            gluCylinder(barrelCylinder, .010, .010, .06, 10, 10);
         } glPopMatrix(); 
      } glPopMatrix();
   } glPopMatrix();
   
}

void display() {
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   
   gluLookAt(eye.x, eye.y, eye.z, look.x, look.y, look.z, 0, 1, 0);
   pos_light();

   if (pointSwt) {
      drawPoints();
   } else {
      drawTriangles();
      //drawBadTriangles();
   }
   if (normalSwt) {
      drawNormals();
      //drawBadNormals();
   }
   drawProjectiles();
   drawTank();
   glutSwapBuffers();
}

int main(int argc, char **argv) {
   //set up my window
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
   glutInitWindowSize(800, 800); 
   glutInitWindowPosition(100, 100);
   glutCreateWindow("Deformable Fractal Terrain");
   glClearColor(0.0, 0.0, 0.0, 0.0);
   GH = GW = 800;
   eyeX = 0.0;
   eyeY = .1;
   eyeZ = -.5;

   lookX = 0.0;
   lookY = 0.0;
   lookZ = 0.0;
   theta = phi = 0;
   moveTrigger = 0;
   buildTrigger = 1;
   pointSwt = normalSwt = wireframe = closest = vehDir = 0;
   flyMode = 0;

   //register glut callback functions
   glutDisplayFunc( display );
   glutReshapeFunc( reshape );
   glutKeyboardFunc(keyboard);
   glutMouseFunc(mouseClick);
   glutPassiveMotionFunc(mouseMotion);
   glutIdleFunc(idle);

   glEnable(GL_DEPTH_TEST);
   glEnable(GL_LIGHTING);
   glShadeModel(GL_SMOOTH);

   if (argc < 4) {
      fprintf(stderr, "Error: Too few arguments\n");
      fprintf(stderr, "Usage: %s <width> <seed> <depth> <range>\n", argv[0]);
      exit(0);
   }
   side = (1 << atoi(argv[3])) + 1;
   length = atof(argv[1]);
   interval = length / side;
   int depth = atoi(argv[3]);
   if (depth == 0) {
      fprintf(stderr, "Depth must be greater than zero\n");
      exit(0);
   }
   TerrGen tGen(length, atoi(argv[2]), depth, atof(argv[4]), &pVec);
   tGen.Generate();
   
   genTriangles();
   genNormals();
  
   // Position Camera
   // Add Movement

   barrelCylinder = gluNewQuadric();
   gluQuadricNormals(barrelCylinder, GLU_SMOOTH);


   
   vehNdx = (side * side - 1) / 2;
   eye = *pVec[vehNdx];
   eye.y += .1;
   //eye.x = 0;
   //eye.y = .1;
   //eye.z = -.5;
   look = eye + Vector(0, 0, -.1);
   init_lighting();
   glutMainLoop();
   return 0;
}

